/**
 * Mesh Analytics SDK
 * @author [Vivek Sudarsan]
 * @version 0.3.137
 */

// Import third party libraries
import Cookies from 'js-cookie';

// Import helpers and util
import apiRequest from './services/api/request';
import { handleJsonResponse } from './services/api/response';
import { UserInteraction } from './interactions';
import {
  getIpAddress,
  getReferrer,
  getUrlParams,
  getScrollPercentage,
  postMessageToAllFrames,
  handleFrameMessage,
  initFormHandlingOnBody,
  avinaLog,
  getBrowserDetails,
  addMeshVisitorIdToForms,
  detectUrlChange,
} from './util/helpers';
import { handleCalendlyEmbed } from './util/integrations/calendly';
import {
  HEARBEAT_INTERVAL,
  INTERACTION_CUSTOMERS_EXCLUDE,
  THIRD_PARTY_SCRIPTS,
  TOUCH_TYPES,
  TRACK_TYPE_MAP,
} from './util/constants';

// version info
const MESH_VERSION = '0.3.137';

// Set type definitions
type Mesh = {
  init?: (key: string, config: any) => void;
  identify?: (contactInfo: Contact) => void;
};

type Contact = {
  firstName?: string;
  lastName?: string;
  ipAddress?: string;
  personId?: string;
  hem?: string;
  maid?: string;
  company?: string;
  title?: string;
  email?: string;
};

// Set global variables
export let glob: any = {
  sdkKey: null,
  sdkAttributes: {
    track: {
      session: true,
      forms: true,
    },
    appMode: false,
  },
  workspaceName: null,
  workspaceId: null,
  workspaceHashSalt: null,
  fullContactKey: null,
  vectorKey: null,
  visitorId: null,
  touchpoints: {},
  fingerprints: {},
  formData: {},
  sessionData: {
    percentPageViewed: 0,
  },
  ongoingRequests: {
    [TOUCH_TYPES.VIEW]: false,
    [TOUCH_TYPES.FORM_FILL]: false,
  },
  settings: {
    sessionHandlingSet: false,
    formHandlingSet: false,
  },
};

declare global {
  interface Window {
    mesh: any;
    meshLogs: any;
    debugInteractions: any;
  }
}

// initialize fingerprint
let fpPromise: any;

/**
 * On load resolve identity and create touchpoint
 */
const initialize = async (config: {
  loadInteractions: boolean;
  loadForms: boolean;
}) => {
  avinaLog('Initializing...');

  // get the sdk key from the script tag
  getScriptAttributes();

  // get corresponding workspace and secrets for sdk key
  await getWorkspaceDetails();

  // load event listeners and third parties
  preLoadEvents();

  // initialize form handling
  initFormHandlingOnBody(onFormFillChange);

  // determine the identity of the visitor
  var identity: any | undefined;
  if (
    glob.sdkAttributes.track === undefined ||
    (glob.sdkAttributes.track &&
      glob.sdkAttributes.track[TRACK_TYPE_MAP.VIEW] !== false)
  ) {
    identity = await resolveIdentity();

    // integration specific handling for visitor
    handleIntegrationFrames();

    // add a touchpoint for this visitor
    await createTouchpoint();
  } else {
    avinaLog('Session tracking disabled...');
  }

  // initialize user interaction tracking
  if (
    !INTERACTION_CUSTOMERS_EXCLUDE.includes(glob.workspaceName) &&
    config.loadInteractions
  ) {
    const userInteraction = new UserInteraction();
    userInteraction.init();
  }

  // add visitor id to all forms on page
  addMeshVisitorIdToForms();

  // handle third party scripts
  await thirdPartyHandling();

  // exit early if in app mode because we don't
  // need to track the session data outside of the touchpoint
  if (glob.sdkAttributes.appMode === true) return;

  // handle session data if session tracking is allowed
  // TODO: combine this with the above if statement
  if (
    glob.sdkAttributes.track === undefined ||
    (glob.sdkAttributes.track &&
      glob.sdkAttributes.track[TRACK_TYPE_MAP.VIEW] !== false)
  )
    handleSessionData();
};

/**
 * Load any event listeners on the page as well as third party scripts
 * @returns
 */
const preLoadEvents = () => {
  // load contact resolution script
  if (glob.contactResolutionEnabled) {
    const vectorScript = THIRD_PARTY_SCRIPTS.vector;
    const vectorScriptElement = document.createElement('script');
    vectorScriptElement.type = 'text/javascript';
    vectorScriptElement.innerHTML = vectorScript;
    document.head.appendChild(vectorScriptElement);
    avinaLog('C resolution loaded...');
  }

  // add event listeners
  window.addEventListener('message', handleFrameMessage, false);
};

/**
 * Get attributes from script to save in global state
 * @returns
 */
const getScriptAttributes = () => {
  // get the sdk key from the script tag
  let sdkScript = document.getElementById('mesh-analytics-sdk');
  glob.sdkKey = sdkScript?.getAttribute('data-mesh-sdk');
  glob.sdkAttributes = JSON.parse(
    sdkScript?.getAttribute('data-mesh-sdk-attributes') || '{}'
  );
};

/**
 * Get and set workspace details for the sdk key
 * @returns
 */
const getWorkspaceDetails = async () => {
  // get corresponding workspace and secrets for sdk key
  const response = await apiRequest({
    method: 'post',
    resource: 'init-sdk',
    id: [],
    data: {
      domain: window.location.href,
      sdkKey: glob.sdkKey,
    },
  });
  const data = await handleJsonResponse(response);

  // assign globals from sdk response
  glob.workspaceId = data.workspaceId;
  glob.workspaceName = data.workspaceName;
  glob.fullContactKey = data.keys.fullContactKey;
  glob.active = data.active;
  glob.contactResolutionEnabled = data.contactResolutionEnabled;
};

/**
 * Handle iFrames for specific integrations
 * ex. Calendly, Drift, etc.
 */
const handleIntegrationFrames = () => {
  // handle calendly frame
  handleCalendlyEmbed(glob.visitorId);
};

/**
 * Check if the visitor identiy exists, if they do
 * return it or create if not
 * @returns visitorId
 */
const resolveIdentity = async () => {
  // check if there's a visitor id cookie on the page
  let visitorId = Cookies.get('mesh-visitor-id');

  // if using fingerprint, get the visitor id
  let personId = null;
  if (glob.sdkAttributes.useFingerprint === true && !visitorId) {
    avinaLog('Cookieless enabled...');

    let fp = await fpPromise;
    let result = await fp.get();
    personId = result.visitorId;

    avinaLog('Person id resolved...');
  }

  // get the hem and maid if exists
  let maid = null;
  let hem = null;

  // gather fingerprints and assign to glob vars
  const fingerprints: any = {
    hem: hem,
    maid: maid,
    personId: personId,
  };

  // get the email of the visitor if in the url parameters
  const urlParams = new URLSearchParams(window.location.search);
  const email = urlParams.get('m_email');
  if (email) fingerprints['email'] = email;

  // assign fingerprints to glob vars
  glob.fingerprints = fingerprints;

  // if there isn't an existing identity, create one
  // in the event of no cookie option, we always do this
  // because we need to verify a contact exists for the
  // person id we've retrieved
  if (!visitorId) visitorId = await createIdentity();

  // assign visitor id to glob vars and return
  avinaLog(`Visitor id resolved... ${visitorId}`);
  glob.visitorId = visitorId;

  // send this visitor id to any embedded frames
  postMessageToAllFrames({ visitorId: visitorId });

  return visitorId;
};

/**
 * Generate a new contact for an unknown visitor
 * @returns visitorId
 */
const createIdentity = async () => {
  // log that we're creating a new identity
  avinaLog('Creating new identity...');

  // create a new contact
  const response = await apiRequest({
    method: 'post',
    resource: 'contacts',
    id: [],
    data: {
      ipAddress: await getIpAddress(),
      workspaceId: glob.workspaceId,
      ...glob.fingerprints,
    },
  });

  // debug: log the new identity
  const data = await handleJsonResponse(response);

  // get the id from the created contact and set cookie
  const visitorId = data.id;
  if (
    glob.sdkAttributes.useFingerprint === false ||
    glob.sdkAttributes.useFingerprint === undefined
  ) {
    avinaLog('Getting visitor id...');
    Cookies.set('mesh-visitor-id', visitorId);
  } else if (glob.sdkAttributes.appMode === true) {
    // when we're in app mode we can store the person id
    avinaLog('App mode enabled...');
    Cookies.set('mesh-visitor-id', visitorId);
  }

  return visitorId;
};

/**
 * Update a contact with new visitor information
 * @returns visitorId
 */
const updateIdentity = async (contactInfo: Contact) => {
  // Check if contactInfo is not provided
  if (!contactInfo) {
    throw new Error('No contact information provided');
  }

  // update an existing contact
  const response = await apiRequest({
    method: 'patch',
    resource: ['contacts'],
    id: [glob.visitorId],
    data: contactInfo,
  });

  // debug: log the updated identity
  const data = await handleJsonResponse(response);

  return data;
};

/**
 * Add a touchpoint for the identified visitor
 * @returns touchpointId
 */
const createTouchpoint = async (type?: string) => {
  // exit early if no visitor id
  if (!glob.visitorId) {
    avinaLog('No visitor id to create touchpoint...');
    throw new Error('[Mesh] No visitor id to create touchpoint');
  }

  // if no type provided default to VIEW
  if (!type) {
    if (glob.sdkAttributes.appMode === true) {
      type = TOUCH_TYPES.IN_APP_VIEW;
    } else {
      type = TOUCH_TYPES.VIEW;
    }
  }

  // if type if form fill, add that to the request
  let formData = { formData: null };
  if (type == TOUCH_TYPES.FORM_FILL) formData = glob.formData;

  // check tracking perferences before writing touchpoint
  if (
    glob.sdkAttributes.track &&
    glob.sdkAttributes.track[TRACK_TYPE_MAP[type]] === false
  )
    return;

  // check if there is an ongoing request for this type
  if (glob.ongoingRequests[type] === true) {
    avinaLog('Ongoing request, skipping...');
    return;
  }

  // set ongoing request for this type
  glob.ongoingRequests[type] = true;

  // create session data
  let sessionData = {
    debug: {
      sdkVersion: MESH_VERSION,
      consoleLogs: window.meshLogs,
      browser: getBrowserDetails(),
    },
  };

  // create a new touchpoint
  const response = await apiRequest({
    method: 'post',
    resource: 'touches',
    id: [],
    data: {
      touchType: type,
      contactId: glob.visitorId,
      pageUrl: window.location.href,
      ipAddress: await getIpAddress(),
      utmParams: await getUrlParams(),
      referrer: await getReferrer(),
      workspaceId: glob.workspaceId,
      formData: JSON.stringify(formData),
      sessionData: JSON.stringify(sessionData),
      ...glob.fingerprints,
    },
  });
  const data = await handleJsonResponse(response);

  // keep track of cur touchpoint ids for each
  // type of touchpoint
  glob.touchpoints[type] = data.id;

  // reset ongoing request for this type
  glob.ongoingRequests[type] = false;

  return data.id;
};

/**
 * Update a current touchpoint with new data or info
 * @returns touchpointId
 */
const updateTouchpoint = async (type: string, updatedTouch: any) => {
  // exit early if no visitor id
  if (!glob.visitorId) {
    avinaLog('No visitor id to update touchpoint...');
    throw new Error('[Mesh] No visitor id to create touchpoint');
  }

  // get touchpoint id for the given type
  const touchpointId = glob.touchpoints[type];
  if (!touchpointId) {
    avinaLog('No touchpoint to update...');
    throw new Error('[Mesh] No touchpoint to update');
  }

  // check tracking perferences before writing touchpoint
  if (
    glob.sdkAttributes.track &&
    glob.sdkAttributes.track[TRACK_TYPE_MAP[type]] === false
  )
    return;

  // check if there is an ongoing request for this type
  if (glob.ongoingRequests[type] === true) {
    avinaLog('Ongoing request, skipping...');
    return;
  }

  // set ongoing request for this type
  glob.ongoingRequests[type] = true;

  // create updated session data
  let sessionData = {
    debug: {
      sdkVersion: MESH_VERSION,
      consoleLogs: window.meshLogs,
      browser: getBrowserDetails(),
    },
  };
  updatedTouch = {
    ...updatedTouch,
    sessionData: JSON.stringify(sessionData),
  };

  // update an existing contact
  const response = await apiRequest({
    method: 'patch',
    resource: ['touches'],
    id: [touchpointId],
    data: updatedTouch,
  });

  // debug: log the updated touchpoint
  const data = await handleJsonResponse(response);

  // reset ongoing request for this type
  glob.ongoingRequests[type] = false;

  return data.id;
};

/**
 * Handle form changes and keep track of the
 * full form data to add to a form fill touchpoint
 */
const onFormFillChange = async (formData: Contact) => {
  // don't update if no formData
  if (Object.keys(formData).length == 0) return;

  // update the information on the contact
  await updateIdentity(formData);

  // update and merge form data
  const curFormData = glob.formData;
  const updatedFormData = Object.assign(curFormData, formData);
  glob.formData = updatedFormData;

  // if there is no form fill touch point for
  // this specific page, create one else update it
  if (!glob.touchpoints[TOUCH_TYPES.FORM_FILL]) {
    createTouchpoint(TOUCH_TYPES.FORM_FILL);
  } else {
    updateTouchpoint(TOUCH_TYPES.FORM_FILL, {
      formData: JSON.stringify({
        ...glob.formData,
      }),
    });
  }
};

/**
 * Handle session data heartbeat
 *
 * Sends data about the session to the server
 * at a set interval so that we can track
 * session activity up till the session ends.
 *
 * ex. send a heartbeat request every 500ms to keep
 * the session duration accurate.
 */
const handleSessionData = async () => {
  // exit early if session handling is already set
  if (glob.settings.sessionHandlingSet) return;

  setInterval(() => {
    // if the window/tab is not active or visible, exit early
    if (document.hidden) return;

    // if there is no view touch point for
    // this specific page, create one else update it
    if (!glob.touchpoints[TOUCH_TYPES.VIEW]) return;

    // calculate session percent viewed as float
    const percentPageViewed = Math.max(
      glob.sessionData.percentPageViewed,
      getScrollPercentage()
    );
    glob.sessionData.percentPageViewed = percentPageViewed;

    // get current session time to record end time
    const currentSessionTime = new Date().toISOString();

    updateTouchpoint(TOUCH_TYPES.VIEW, {
      sessionEndedAt: currentSessionTime,
      sessionPercentViewed: percentPageViewed,
    });
  }, HEARBEAT_INTERVAL);

  // set session handling set to true
  glob.settings.sessionHandlingSet = true;
};

const thirdPartyHandling = async () => {
  let attempts = 0;
  const vectorIdentify = async () => {
    try {
      // @ts-ignore
      if (typeof vector === 'undefined' && attempts < 5) {
        avinaLog('Third party script identify not loaded, retrying...');
        attempts++;
        setTimeout(vectorIdentify, 2500);
        return;
      }

      // @ts-ignore
      await vector?.load('c80d8374-fa6c-444a-b596-2869133db0f7');

      // @ts-ignore
      await vector?.identify(
        JSON.stringify({
          contact_id__mesh_web: glob.visitorId,
          workspace_id: glob.workspaceId,
        })
      );
    } catch (e) {
      avinaLog(`Third party script identify failed...`);
    }
  };

  // Run third party identify scripts
  await vectorIdentify();
};

/**
 * IIFE that runs on load and attach
 * global mesh functions to window
 */
(function() {
  var mesh: Mesh = {};

  // initialize sdk
  initialize({
    loadInteractions: true,
    loadForms: true,
  });

  // sdk function
  mesh.init = async (key, config) => {};

  // expose ability to write user data
  mesh.identify = async (contactInfo: Contact) => {
    updateIdentity(contactInfo);
  };

  // Wait for the window to load before
  // making the SDK available globally
  window.onload = function() {
    // define the namespace mesh
    window.mesh = mesh;
  };

  // listen for popstate events to reinitialize
  // and run url change detection to handle the case
  // where it's an SPA and the page doesn't reload
  // when the url changes
  detectUrlChange(initialize);
  window.addEventListener('popstate', () => {
    avinaLog('URL changed, reinitializing...');
    initialize({
      loadInteractions: false,
      loadForms: false,
    });
  });
})();
